from scipy.sparse.linalg import lsqr

def sparse_lsqr_solver(A, b):
    x, istop, itn, r1norm, r2norm, *args = lsqr(A, b)
    print('SOLVER LOG istop:', istop, 'itn:', itn, 'r1norm:', r1norm, 'r2norm:', r2norm)
    return x




from .utils import MultiDimLinearExpr, multidim_lin_expr, linear_expr

from collections import defaultdict
import networkx as nx
import numpy as np


class GaussianStitcher(object):
    def __init__(self, n_dims, solver_hook):
        self.n_dims = n_dims
        self.solver_hook = solver_hook
    
    def stitch(self, data_in, v_origin):
        digraph = self._make_digraph(data_in)
        lin_expressions = self._make_constraints(digraph, v_origin)
        node2coordinates = self._optimize(lin_expressions)
        return node2coordinates, digraph

    def _make_digraph(self, data_in):
        digraph = nx.DiGraph()
        for v, w, Lambda_vw, p_vw in data_in:
            assert Lambda_vw.shape == (self.n_dims, self.n_dims)
            assert p_vw.shape == (self.n_dims,)
            digraph.add_edge(v, w, Lambda=Lambda_vw, p=p_vw)
        return digraph
    
    def _make_constraints(self, digraph, v_origin):
        # ADD GAUSSIAN POTENTIAL
        lin_expressions = MultiDimLinearExpr([])
        for u in digraph.nodes():
            multidim_lin_expr_acc = MultiDimLinearExpr([])
            for v in digraph.nodes():
                for w in digraph.successors(v):
                    if v == w:
                        raise ValueError('Self loops are not allowed. Check vertex {}.'.format(v))
                    if u == v or u == w:
                        delta_vwu = 1. if u == v else -1.
                        p_vw = digraph[v][w]['p']
                        Lambda_vw = digraph[v][w]['Lambda']
                        current_lin_expr = multidim_lin_expr(
                            variables = [('s', v, w, dim_i) for dim_i in range(self.n_dims)],
                            A = delta_vwu * Lambda_vw,
                            b = - delta_vwu * np.dot(Lambda_vw, p_vw)
                        )
                        multidim_lin_expr_acc = multidim_lin_expr_acc.add(current_lin_expr)
                    else:
                        pass
            lin_expressions.extend(multidim_lin_expr_acc)
    
        # ADD CONSTRAINTS ON SLACKS
        for v, w in digraph.edges():
            for dim_i in range(self.n_dims):
                lin_expressions.append(
                    linear_expr(
                        variables = [
                            ('s', v, w, dim_i),
                            ('t', v, dim_i),
                            ('t', w, dim_i)
                        ],
                        coefficients = [
                            1.,
                            -1.,
                            1.
                        ]
                    )
                )
        # ADD CONSTRAINT FOR THE ORIGIN
        if not digraph.has_node(v_origin):
            raise ValueError('Origin node {} is missing!!'.format(v_origin))

        for dim_i in range(self.n_dims):
            lin_expressions.append(
                linear_expr(
                    variables = [('t', v_origin, dim_i)],
                    coefficients = [1.]
                )
            )
        return lin_expressions

    def _optimize(self, lin_expressions):
        variables, A, b = lin_expressions.get_vars_and_matrices()
        x = self.solver_hook(A, b)
        var2value = {vi:xi for vi, xi in zip(variables, x)}
        node2coordinates = get_node2coordinates(var2value, n_dims=self.n_dims)
        return node2coordinates


def get_node2coordinates(var2value, n_dims):
    node2coordinates = defaultdict(lambda:np.zeros((n_dims,)))
    for var, value in var2value.items():
        if var[0] == 't':
            _, v, dim_i = var
            node2coordinates[v][dim_i] = value
    return node2coordinates



from .mapper import QPBuilder
from qpsolvers import solve_qp

from collections import defaultdict
import networkx as nx
import numpy as np


def get_node2coordinates(x, nodes, n_dims):
    node2coordinates = {}
    for idx, v in enumerate(nodes):
        node2coordinates[v] = x[idx*n_dims: (idx+1)*n_dims]
    return node2coordinates


class GaussianQPBuilder(object):
    def __init__(self, n_dims, digraph, v_origin):
        self.n_dims = n_dims
        self.digraph = digraph
        self.v_origin = v_origin
        self.builder = QPBuilder(dtype=np.double)
        self.node2coords = {
            v: self.builder.new_vector((v,), self.n_dims)
                for v in digraph.nodes()
        }
        self.edge2coords = {
            (v, w): self.builder.new_vector((v, w), self.n_dims)
                for v in digraph.nodes()
                    for w in digraph.successors(v)
        }

    def set_objective(self):
        objective = self.builder.new_objective()
        for v in self.digraph.nodes():
            for w in self.digraph.successors(v):
                data = self.digraph[v][w]
                Lambda_vw = data['Lambda']
                s_vw = self.edge2coords[(v, w)]
                objective.add_quad(s_vw, Lambda_vw, s_vw)

    def add_diff_equalities(self):
        for v in self.digraph.nodes():
            t_v = self.node2coords[v]
            for w in self.digraph.successors(v):
                t_w = self.node2coords[w]

                p_vw = self.digraph[v][w]['p']
                s_vw = self.edge2coords[(v, w)]
                # DIFF EQUALITIES
                for dim in range(self.n_dims):
                    diff_eq = self.builder.new_equality()
                    diff_eq.add_dotprod([t_w[dim], t_v[dim], p_vw[dim], s_vw[dim]], [1, -1, -1, -1])

    def add_orig_equalities(self):
        t_orig = self.node2coords[self.v_origin]
        for dim in range(self.n_dims):
            origin_eq = self.builder.new_equality()
            origin_eq.add_dotprod([t_orig[dim]], [1])

    def add_ub_inequalities(self):
        for v in self.digraph.nodes():
            t_v = self.node2coords[v]
            for w in self.digraph.successors(v):
                t_w = self.node2coords[w]
                ub = self.digraph[v][w].get('ub', None)
                if ub is not None:
                    for dim in range(self.n_dims):
                        ub_ineq = self.builder.new_inequality()
                        ub_ineq.add_dotprod([t_w[dim], t_v[dim], ub[dim]], [1, -1, -1])

    def add_lb_inequalities(self):
        for v in self.digraph.nodes():
            t_v = self.node2coords[v]
            for w in self.digraph.successors(v):
                t_w = self.node2coords[w]
                lb = self.digraph[v][w].get('lb', None)
                if lb is not None:
                    for dim in range(self.n_dims):
                        # lb <= t_w-t_v
                        # t_v-t_w + lb <= 0
                        lb_ineq = self.builder.new_inequality()
                        lb_ineq.add_dotprod([t_v[dim], t_w[dim], lb[dim]], [1, -1, 1])


    def build(self):
        return self.builder.build()

    def variables(self):
        return self.builder.variables()



class GaussianStitcherQP(object):
    def __init__(self, n_dims, solver):
        self.n_dims = n_dims
        self.solver = solver

    def stitch(self, data_in, v_origin):
        digraph = self._make_digraph(data_in)
        x, variables = self._optimize(digraph, v_origin)
        if x is None:
            raise Exception("solver error")
        x_sol = np.array(x).reshape((-1,))
        # print('VARS', variables)
        # print('VALS', x_sol)
        node2coords, _edge2coords = self._sol2coords(x_sol, variables)
        return node2coords, digraph

    def _sol2coords(self, x_sol, variables):
        assert len(variables) == len(x_sol)
        node2coords = defaultdict(lambda:np.zeros((self.n_dims,)))
        edge2coords = defaultdict(lambda:np.zeros((self.n_dims,)))
        for variable, value in zip(variables, x_sol):
            elem, coord_idx = variable.name
            if len(elem) == 1:
                node2coords[elem[0]][coord_idx] = value
            elif len(elem) == 2:
                edge2coords[elem][coord_idx] = value
            else:
                raise ValueError
        return dict(node2coords), dict(edge2coords)
#
    def _make_digraph(self, data_in):
        digraph = nx.DiGraph()
        for constraint_tuple in data_in:
            digraph.add_edge(
                constraint_tuple.v,
                constraint_tuple.w,
                Lambda=constraint_tuple.Lambda,
                p=constraint_tuple.p,
                lb=constraint_tuple.lb,
                ub=constraint_tuple.ub
            )
        return digraph
#
    def get_matrices(self, digraph, v_origin):
        gbuilder = GaussianQPBuilder(self.n_dims, digraph, v_origin)
        gbuilder.set_objective()
        gbuilder.add_diff_equalities()
        gbuilder.add_orig_equalities()
        gbuilder.add_lb_inequalities()
        gbuilder.add_ub_inequalities()
        solver_matrices = gbuilder.build()
        return solver_matrices, gbuilder.variables()

#
    def _optimize(self, digraph, v_origin):
        solver_matrices, variables = self.get_matrices(digraph, v_origin)
        P = solver_matrices.P
        q = solver_matrices.q
        G = solver_matrices.G
        h = solver_matrices.h
        A = solver_matrices.A
        b = solver_matrices.b

        # PAG  = np.concatenate([P, A, G])
        # print('rank(A)', np.linalg.matrix_rank(A))
        # print('shape(A)', A.shape)
        # print('rank([P; A; G])', np.linalg.matrix_rank(PAG))
        # print('shape([P; A; G])', PAG.shape)

        x = solve_qp(P, q, G, h, A, b, solver=self.solver)
        return x, variables
#



import math
import logging
import os.path

from functools import lru_cache

import psutil
import numpy as np
import tifffile as tiff

from zetastitcher import VirtualFusedVolume


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class FuseRunner(object):
    def __init__(self, file_matrix):
        self.fm = file_matrix  #: :class:`FileMatrix`
        self.path, _ = os.path.split(self.fm.input_path)

        self.zmin = 0
        self.zmax = None
        self.downsample_xy = None
        self.output_filename = None
        self.channel = None
        self.compression = 'zlib'

        self._is_multichannel = None

        self.vfv = VirtualFusedVolume(file_matrix)

    @property
    @lru_cache()
    def dtype(self):
        return self.vfv.dtype

    @property
    def debug(self):
        return self.vfv.overlay_debug_enabled

    @debug.setter
    def debug(self, value):
        self.vfv.overlay_debug_enabled = value

    @property
    def is_multichannel(self):
        if self.channel is not None:
            return False

        return self.vfv.nchannels > 1

    @property
    def output_shape(self):
        output_shape = list(self.vfv.shape)

        thickness = output_shape[0]
        if self.zmax is not None:
            thickness -= (thickness - self.zmax)
        thickness -= self.zmin

        output_shape[0] = thickness

        if self.downsample_xy:
            output_shape[-2] /= self.downsample_xy
            output_shape[-1] /= self.downsample_xy

        if self.vfv.nchannels > 1 and not self.is_multichannel:
            del output_shape[1]

        return tuple(map(math.ceil, output_shape))

    def run(self):
        total_byte_size = (np.prod(self.output_shape) * self.dtype.itemsize).item()
        bigtiff = total_byte_size > 2**31 - 1

        ram = psutil.virtual_memory().available

        # size in bytes of an xy plane (including channels) (float32)
        xy_size = (np.prod(self.output_shape[1::]) * 4).item()
        n_frames_in_ram = int(ram / xy_size / 2.5)

        n_loops = self.output_shape[0] // n_frames_in_ram

        partial_thickness = [n_frames_in_ram for _ in range(0, n_loops)]
        remainder = self.output_shape[0] % n_frames_in_ram
        if remainder:
            partial_thickness += [remainder]

        try:
            os.remove(self.output_filename)
        except FileNotFoundError:
            pass

        for thickness in partial_thickness:
            self.zmax = self.zmin + thickness
            ie = [
                slice(self.zmin, self.zmax),
                Ellipsis if self.channel is None else self.channel,
            ]
            if self.downsample_xy:
                ie += [slice(None, None, self.downsample_xy)] * 2

            fused = self.vfv[ie]

            if self.is_multichannel:
                fused = np.moveaxis(fused, -3, -1)

            logger.info('saving output to {}'.format(self.output_filename))
            tiff.imwrite(self.output_filename, fused, append=True, bigtiff=bigtiff, compression=self.compression)

            self.zmin += thickness



import logging

import numpy as np
import pandas as pd

from zetastitcher.gaussian_stitcher.qp.stitching import GaussianStitcherQP as GaussianStitcher
from zetastitcher.gaussian_stitcher.common import ConstraintTuple


logger = logging.getLogger(__name__)


def global_optimization(df, xcfm):
    """Perform global optimization to adjust tile absolute positions.

    The passed dataframe must come with an initial guess of absolute positions.

    Parameters
    ----------
    df : :attr:`.FileMatrix.data_frame`
    sdf : :attr:`.FileMatrix.stitch_data_frame`
    """
    logger.info('performing global optimization...')
    N_DIMS = 3
    CTND = ConstraintTuple(N_DIMS)
    Eye3 = np.eye(N_DIMS)

    data_in = []

    xcorr_options = xcfm.xcorr_options
    max_dx = xcorr_options['max_dx']
    max_dy = xcorr_options['max_dy']
    max_dz = xcorr_options['max_dz']
    overlap_h = xcorr_options['overlap_h']
    overlap_v = xcorr_options['overlap_v']

    sdf = xcfm.stitch_data_frame

    for idx, (src, row) in enumerate(sdf.iterrows()):
        dst = row['bname']
        score = row['score']
        score = np.float32(score)
        p = np.array([row['px'], row['py'], row['pz']], dtype=np.float32)
        src_tile = df.loc[src]
        if row['axis'] == 1:
            lb = np.array([-max_dx, src_tile['ysize'] - overlap_v - max_dy, -max_dz])
            ub = np.array([max_dx, src_tile['ysize'] - overlap_v + max_dy, max_dz])
        elif row['axis'] == 2:
            lb = np.array([src_tile['xsize'] - overlap_h - max_dy, -max_dx, -max_dz])
            ub = np.array([src_tile['xsize'] - overlap_h + max_dy, max_dx, max_dz])
        else:
            raise ValueError('invalid axis {}'.format(row['axis']))
        data_in.append(CTND(src, dst, Eye3 * score, p, lb=lb, ub=ub))

    cond = (df['X'] == 0) & (df['Y'] == 0) & (df['Z'] == 0)
    v_origin = df[cond].index[0]  # top-left tile

    stitcher = GaussianStitcher(
        n_dims=N_DIMS,
        solver='cvxpy'
    )
    node2coordinates, digraph = stitcher.stitch(data_in, v_origin)

    laplacian_df = pd.DataFrame(node2coordinates).transpose()

    keys = ['Xs', 'Ys', 'Zs']  # do not change order
    df[keys] = laplacian_df
    df[keys] -= df[keys].min()
    df[keys] = df[keys].apply(np.round).astype(int)
    df['Xs_end'] = df['Xs'] + df['xsize']
    df['Ys_end'] = df['Ys'] + df['ysize']
    df['Zs_end'] = df['Zs'] + df['nfrms']


def compute_shift_vectors(filematrix_dataframe, stitch_dataframe):
    sdf = stitch_dataframe
    fm_df = filematrix_dataframe

    sdf['px'] = 0
    sdf['py'] = 0
    sdf['pz'] = 0

    idx = sdf['axis'] == 1
    fm_df_idx = idx[idx].index
    sdf.loc[idx, 'px'] = sdf.loc[idx, 'dx']
    sdf.loc[idx, 'py'] = fm_df.loc[fm_df_idx, 'ysize'] - sdf.loc[idx, 'dy']
    sdf.loc[idx, 'pz'] = sdf.loc[idx, 'dz']

    idx = sdf['axis'] == 2
    fm_df_idx = idx[idx].index
    sdf.loc[idx, 'px'] = fm_df.loc[fm_df_idx, 'xsize'] - sdf.loc[idx, 'dy']
    sdf.loc[idx, 'py'] = -sdf.loc[idx, 'dx']
    sdf.loc[idx, 'pz'] = sdf.loc[idx, 'dz']



import yaml
import pandas as pd


class XcorrFileMatrix:
    def __init__(self):
        self.xcorr_options = None

        self.stitch_data_frame = None

        self.ascending_tiles_x = None
        self.ascending_tiles_y = None

    @classmethod
    def from_data(cls, xcorr_options, xcorr_df):
        inst = cls()

        attrs = ['ascending_tiles_x', 'ascending_tiles_y']

        for attr in attrs:
            setattr(inst, attr, xcorr_options[attr])

        stitch_data_frame = xcorr_df
        stitch_data_frame = stitch_data_frame.rename(columns={'aname': 'filename'})
        stitch_data_frame = stitch_data_frame.set_index('filename')

        inst.stitch_data_frame = stitch_data_frame
        inst.xcorr_options = xcorr_options

        return inst

    @classmethod
    def from_yaml(cls, fname):
        with open(fname, 'r') as f:
            y = yaml.safe_load(f)

        return XcorrFileMatrix.from_data(y['xcorr-options'], pd.DataFrame(y['xcorr']))

    def aggregate_results(self):
        sdf = self.stitch_data_frame.reset_index()

        view = sdf.loc[sdf.groupby(['filename', 'bname', 'axis']).score.idxmax()]
        overlap_dict = {1: self.xcorr_options['overlap_v'],
                        2: self.xcorr_options['overlap_h']}

        view.dz -= self.xcorr_options['max_dz']
        for a in [1, 2]:
            indexes = (view['axis'] == a)
            view.loc[indexes, 'dy'] = (overlap_dict[a]
                                       + self.xcorr_options['max_dy']
                                       - view.loc[indexes, 'dy'])
        view.dx -= self.xcorr_options['max_dx']

        self.stitch_data_frame = view.set_index('filename')



try:
    # Suppress warnings when importing pims on a machine without display server
    # The warnings are due to import matplotlib.pyplot as plt in pim's display.py
    import matplotlib as mpl
    mpl.use('Agg')
except ModuleNotFoundError:
    pass

from pathlib import Path
from pims import ImageSequence

from zetastitcher.io.inputfile_mixin import InputFileMixin


class PimsWrapper(InputFileMixin):
    def __init__(self, initializer=None):
        super().__init__()

        self.initializer = initializer
        self.imseq = None

        if self.initializer is not None:
            self.open()

    def open(self, initializer: Path=None):
        if initializer is not None:
            self.initializer = initializer

        if not self.initializer.is_dir():
            raise ValueError(f'{self.initializer} must be a directory')

        self.imseq = ImageSequence(str(self.initializer))
        setattr(self, 'close', getattr(self.imseq, 'close'))

        self.nfrms = len(self.imseq)
        self.ysize = self.imseq.frame_shape[0]
        self.xsize = self.imseq.frame_shape[1]
        self.dtype = self.imseq._dtype

        if len(self.imseq.frame_shape) > 2:
            self.nchannels = self.imseq.frame_shape[2]

    def frame(self, index, dtype=None):
        a = self.imseq[index]
        if dtype is not None:
            a = a.astype(dtype)
        return a



from pathlib import Path

import numpy as np
import tifffile as tiff

from zetastitcher.io.inputfile_mixin import InputFileMixin


class TiffWrapper(InputFileMixin):
    def __init__(self, file_path=None):
        super().__init__()
        self.file_path = file_path

        self.tfile = None
        self.flist = None
        self.glob_mode = False
        self.axes = None

        if file_path is not None:
            self.file_path = Path(file_path)
            self.open()

    def open(self, file_path=None):
        if file_path is not None:
            self.file_path = Path(file_path)

        if self.file_path.is_dir():
            self.glob_mode = True
            flist = []
            flist += list(Path(self.file_path).glob('*.tif*'))
            flist += list(Path(self.file_path).glob('*.TIF*'))
            flist = sorted(flist)
            fname = flist[0]
            self.flist = flist
        else:
            self.glob_mode = False
            fname = self.file_path

        self.tfile = tiff.TiffFile(str(fname))
        setattr(self, 'close', getattr(self.tfile, 'close'))

        self.dtype = np.dtype(self.tfile.pages[0].dtype)
        self.xsize = self.tfile.pages[0].imagewidth
        self.ysize = self.tfile.pages[0].imagelength

        axes = self.tfile.series[0].axes

        if 'C' in axes:
            self.nchannels = self.tfile.series[0].shape[axes.index('C')]
        elif 'S' in axes:
            self.nchannels = self.tfile.series[0].shape[axes.index('S')]

        if 'Z' in axes:
            nfrms = self.tfile.series[0].shape[axes.index('Z')]
        elif not 'C' in axes:
            nfrms = len(self.tfile.pages)
        else:
            nfrms = 1
        if self.glob_mode:
            nfrms *= len(self.flist)

        self.nfrms = nfrms
        self.axes = axes

    def zslice(self, arg1, arg2=None, step=1, dtype=None, copy=True):
        myslice = self._args_to_slice(arg1, arg2, step)

        if self.glob_mode:
            flist = self.flist[myslice]
            if not flist:
                return np.array([])
            a = tiff.imread(list(map(str, flist)))
        else:
            mma = self.tfile.asarray(out='memmap')
            a = mma

            for letter in 'ZQ':
                if letter in self.axes:
                    # take only first element of left-most axes (e.g. time)
                    mma = mma[tuple(0 for _ in range(self.axes.index(letter)))]
                    if len(mma.shape) >= 3:
                        a = mma[myslice]
                    break

            if copy:
                a = np.copy(a)

        if 'C' in self.axes:
            a = np.moveaxis(a, self.axes.index('C') - len(self.axes), -1)

        if dtype is None:
            return a
        return a.astype(dtype)



import numpy as np
from collections import namedtuple, defaultdict, Counter
import numbers
SolverMatrices = namedtuple('SolverMatrices', 'P, q, G, h, A, b')


def add_sv(d1, d2):
    d = defaultdict(float)
    for k, v in d1.items(): d[k] += v
    for k, v in d2.items(): d[k] += v
    return dict(d)

class Variable(object):
    def __init__(self, factory, name):
        self.factory = factory
        self.name = name

    def __eq__(self, other):
        return self.factory is other.factory and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        return str(self.name).__lt__(str(other.name))

    def __repr__(self):
        return str(self.name)

    def __str__(self):
        try:
            name, coords = self.name
            return "{}_{}".format('.'.join(name), coords)
        except:
            return  str(self.name)


class Term(object):
    def __init__(self, variables):
        self.variables = tuple(sorted(variables))

    def order(self):
        return len(self.variables)

    def __hash__(self):
        return hash(self.variables)

    def __eq__(self, other):
        return self.variables == other.variables

    def __lt__(self, other):
        assert isinstance(other, Term)
        self_order, other_order = self.order(), other.order()
        return self_order.__le__(other_order) or (self_order == other_order and self.variables.__le__(other.variables))

    def __repr__(self):
        return " ".join(map(str, sorted(self.variables)))

    def __str__(self):
        counter = Counter(self.variables)
        l = []
        for k in sorted(counter):
            exponent = counter[k]
            if exponent > 1:
                l.append("{}^{}".format(str(k), exponent))
            else:
                assert exponent == 1
                l.append(str(k))
        return " ".join(map(str, l))


class Expression(object):
    def __init__(self, factory, term2coeff=None):
        self.factory = factory
        if term2coeff is None:
            self.term2coeff = {}
        else:
            self.term2coeff = dict(term2coeff)

    def zero(self):
        return Expression(self.factory, {})

    def one(self):
        return Expression(self.factory, {Term(()):1.})

    def __mul__(self, elem):
        if isinstance(elem, numbers.Number):
            term2coeff = {t:c*elem for t, c in self.term2coeff.items()}
        elif isinstance(elem, Variable):
            term2coeff = {Term(t.variables + (elem,)):c for t, c in self.term2coeff.items()}
        else:
            raise ValueError("elem has type {} ----> {}".format(
                str(type(elem)),
                str(elem.shape)
            ))
        return Expression(self.factory, term2coeff)

    def __imul__(self, other):
        self.term2coeff = self.__mul__(other).term2coeff
        return  self

    def __add__(self, other):
        if self.factory is not other.factory: raise ValueError
        if isinstance(other, Expression):
            return Expression(self.factory, term2coeff=add_sv(self.term2coeff, other.term2coeff))
        elif isinstance(other, Variable):
            return Expression(self.factory, term2coeff=add_sv(self.term2coeff, {Term((other,)):1.}))
        elif isinstance(other, Term):
            return Expression(self.factory, term2coeff=add_sv(self.term2coeff, {other:1.}))

    def __iadd__(self, other):
        self.term2coeff = self.__add__(other).term2coeff
        return self

    def __repr__(self):
        return ' '.join("{:+} {}".format(coeff, term) for term, coeff in sorted(self.term2coeff.items(), key=lambda x:x[0]) if coeff != 0)

    def __str__(self):
        l = []
        for term, coeff in sorted(self.term2coeff.items(), key=lambda x:x[0]):
            if coeff != 0:
                if coeff == 1. and term.order() > 0:
                    l.append("+ {}".format(str(term)))
                else:
                    l.append("{:+} {}".format(coeff, str(term)))
        return ' '.join(l)

    def order(self):
        if self.term2coeff:
            return max([term.order() for term in self.term2coeff])
        else:
            return 0

    def add_term(self, term, coeff):
        assert isinstance(term, Term)
        for v in term.variables:
            self.factory.add_variable(v)

        if term not in self.term2coeff:
            self.term2coeff[term] = coeff
        else:
            self.term2coeff[term] += coeff

    def add_dotprod(self, x, z):
        for xi, zi in zip(x, z):
            elem_expr = self.one()
            elem_expr *= xi
            elem_expr *= zi
            self += elem_expr

    def add_quad(self, vars_v, coeff_mat, vars_w):
        n, m = len(vars_v), len(vars_w)
        assert coeff_mat.shape == (n, m)
        for i, v_i in enumerate(vars_v):
            for j, w_j in enumerate(vars_w):
                elem_expr = self.one()
                elem_expr *= coeff_mat[i, j]
                elem_expr *= v_i
                elem_expr *= w_j
                self += elem_expr

class QPBuilder(object):
    def __init__(self, dtype):
        self.dtype = dtype
        self.objective = None
        self.inequality_list = []
        self.equality_list = []
        self.var_set = set()

    def print_(self):
        return
        print('objective')
        print(self.objective)
        print('s.t.')
        for eq in self.equality_list:
            print(eq, '== 0')
        for ineq in self.inequality_list:
            print(ineq, '<= 0')

    def new_objective(self):
        if self.objective is not None:
            raise ValueError
        self.objective = self._create_expression()
        return self.objective

    def new_variable(self, name):
        variable = Variable(self, name)
        self.var_set.add(variable)
        return variable

    def var2idx(self):
        return {v:idx for idx, v in enumerate(self.variables())}

    def variables(self):
        return sorted(self.var_set)

    def add_variable(self, v):
        self.var_set.add(v)

    def _create_expression(self):
        return Expression(factory=self)

    def new_equality(self):
        expr = self._create_expression()
        self.equality_list.append(expr)
        return expr

    def new_inequality(self):
        # print('DEBUG INEQUALITIES')
        expr = self._create_expression()
        self.inequality_list.append(expr)
        return expr

    def new_vector(self, name, ndims):
        return [self.new_variable((name, i)) for i in range(ndims)]

    def Qp_matrices(self, variables, var2idx):
        n = len(variables)
        Q = np.zeros((n, n), dtype=self.dtype)
        p = np.zeros((n,), dtype=self.dtype)
        for term, coeff in self.objective.term2coeff.items():
            order = term.order()
            if order == 0:
                pass # ignore constants in the objective
            elif order == 1:
                x, = term.variables
                i_x = var2idx[x]
                p[i_x] += coeff
            elif order == 2:
                x, y = term.variables
                i_x = var2idx[x]
                i_y = var2idx[y]
                Q[i_x, i_y] += 2*coeff # XXX x2 because it will be divided by 2 in the objective
            else:
                raise ValueError('Term order higher than 2. TERM {}'.format(term))
        return Q, p

    def Ab_matrices(self, variables, var2idx):
        # print('Ab_matrices\n', '\n\t'.join(map(str, self.equality_list)))
        A, b = self.lin_constr2arrays(self.equality_list, variables, var2idx)
        return A, b

    def Gh_matrices(self, variables, var2idx):
        # print('Gh_matrices\n', '\n\t'.join(map(str, self.inequality_list)))
        G, h = self.lin_constr2arrays(self.inequality_list, variables, var2idx)
        return G, h

    def build(self):
        variables = self.variables()
        var2idx = self.var2idx()
        self.print_()
        # print('DEBUG variables', variables)
        # print('DEBUG var2idx', var2idx)
        P, q = self.Qp_matrices(variables, var2idx)
        A, b = self.Ab_matrices(variables, var2idx)
        G, h = self.Gh_matrices(variables, var2idx)
        return SolverMatrices(P, q, G, h, A, b)


    def lin_constr2arrays(self, constr_list, variables, var2idx):
        A_list = []
        b_list = []
        nvars = len(variables)
        for expr in constr_list:
            if expr.order() != 1:
                raise ValueError('Order {} constraints are not allowed {}'.format(expr.order(), str(expr)))
            row = np.zeros((nvars,), dtype=self.dtype)
            b = 0.
            for term, coeff in expr.term2coeff.items():
                if term.order() == 0:
                    b -= coeff
                elif term.order() == 1:
                    variable, = term.variables
                    vidx = var2idx[variable]
                    row[vidx] += coeff
                else:
                    raise ValueError
            A_list.append(row.reshape((1, -1)))
            b_list.append(b)
        if A_list:
            return np.concatenate(A_list), np.array(b_list, dtype=self.dtype)
        else:
            return None, None



major = 0
minor = 7
micro = 0

pre_release = None
post_release = 1
dev_release = None

__version__ = ''

if major is not None:
    __version__ += f'{major}'

if minor is not None:
    __version__ += f'.{minor}'

if micro is not None:
    __version__ += f'.{micro}'

if pre_release is not None:
    __version__ += f'{pre_release}'

if post_release is not None:
    __version__ += f'.post{post_release}'

if dev_release is not None:
    __version__ += f'.dev{dev_release}'



import mmap
from pathlib import Path

import numpy as np

from zetastitcher.io.inputfile_mixin import InputFileMixin


class MHDWrapper(InputFileMixin):
    def __init__(self, file_path=None):
        super().__init__()
        self.file_path = file_path
        self.raw_file_name = None

        self.a = None
        self.mm = None
        self.file = None
        self.nchannels = 1

        if file_path is not None:
            self.file_path = Path(self.file_path)
            self.open()

    @staticmethod
    def element_type_to_type_fmt(et):
        if not et.startswith('MET_'):
            raise ValueError('Invalid ElementType')

        _, eltype = et.split('MET_')

        type_fmt = None
        if eltype.endswith('CHAR'):
            byte_depth = 1
        elif eltype.endswith('SHORT'):
            byte_depth = 2
        elif eltype.endswith('INT'):
            byte_depth = 4
        elif eltype.endswith('LONG'):
            byte_depth = 8
        elif eltype == 'FLOAT':
            type_fmt = 'f'
            byte_depth = 4
        elif eltype == 'DOUBLE':
            type_fmt = 'f'
            byte_depth = 8
        else:
            raise ValueError('Invalid ElementType')

        if type_fmt is None:
            if eltype.startswith('U'):
                type_fmt = 'u'
            else:
                type_fmt = 'i'

        return '{}{}'.format(type_fmt, byte_depth)

    def open(self, file_name=None):
        if file_name is not None:
            self.file_path = file_name

        d = {}

        with self.file_path.open('r') as f:
            try:
                for line in f.readlines():
                    k, v = map(str.strip, line.split('='))
                    d[k] = v
            except UnicodeDecodeError:
                raise ValueError('Invalid file type')

        if d['ObjectType'] != 'Image':
            raise ValueError('Invalid ObjectType')
        if d['BinaryData'] != 'True':
            raise ValueError('BinaryData must be True')

        type_fmt = self.element_type_to_type_fmt(d['ElementType'])

        byte_order = ''
        if 'BinaryDataByteOrderMSB' in d:
            if d['BinaryDataByteOrderMSB'] == 'True':
                byte_order = '>'
            elif d['BinaryDataByteOrderMSB'] == 'False':
                byte_order = '<'
            else:
                raise ValueError('Invalid BinaryDataByteOrderMSB')

        type_str = '{}{}'.format(byte_order, type_fmt)
        self.dtype = np.dtype(type_str)
        byte_depth = self.dtype.itemsize

        dims = d['DimSize'].split(' ')
        if len(dims) != int(d['NDims']) != 3:
            raise ValueError('Invalid number of dimensions')

        self.xsize, self.ysize, self.nfrms = map(int, dims)

        strides = [
            self.ysize * self.xsize * byte_depth,
            self.xsize * byte_depth,
            byte_depth
        ]

        rawfile = self.file_path.parent.joinpath(d['ElementDataFile'])
        self.file = rawfile.open('r')
        self.mm = mmap.mmap(self.file.fileno(), 0, access=mmap.ACCESS_COPY)
        self.a = np.ndarray(self.shape, self.dtype, self.mm, 0, strides)
        setattr(self, '__getitem__', getattr(self.a, '__getitem__'))

    def close(self):
        if self.mm is not None:
            self.mm.close()
        del self.mm
        self.mm = None
        if self.file is not None:
            self.file.close()

    def zslice(self, arg1, arg2=None, step=None, dtype=None, copy=True):
        a = self.a[self._args_to_slice(arg1, arg2, step)]
        if dtype is not None:
            return a.astype(dtype)
        if copy:
            a = np.copy(a)
        return a



import logging
import argparse
from pathlib import Path

import yaml
import coloredlogs

import numpy as np

from zetastitcher import VirtualFusedVolume, FileMatrix
from zetastitcher.align.xcorr_filematrix import XcorrFileMatrix


logger = logging.getLogger(__name__)
coloredlogs.install(level='INFO', fmt='%(levelname)s [%(name)s]: %(message)s')


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def parse_args():
    parser = argparse.ArgumentParser(
        description='Downscale coordinates in a stitch.yml file.',
        epilog='Author: Giacomo Mazzamuto <mazzamuto@lens.unifi.it>',
        formatter_class=CustomFormatter)

    parser.add_argument('input_file', help='input yml file')
    parser.add_argument('output_file', help='output yml file')

    parser.add_argument('--xy-divide-by', type=float)
    parser.add_argument('--z-divide-by', type=float)
    parser.add_argument('--ext', type=str, default='.tiff',
                        help='replace tile extension (e.g. .tiff)')

    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    fm = FileMatrix(args.input_file)
    df = fm.data_frame

    df['Xs'] /= args.xy_divide_by
    df['Ys'] /= args.xy_divide_by
    df['Zs'] /= args.z_divide_by

    df['xsize'] /= args.xy_divide_by
    df['ysize'] /= args.xy_divide_by
    df['nfrms'] /= args.z_divide_by

    keys = ['Zs', 'Ys', 'Xs', 'xsize', 'ysize', 'nfrms']
    df[keys] = df[keys].apply(np.round).astype(int)

    df['Xs_end'] = df['Xs'] + df['xsize']
    df['Ys_end'] = df['Ys'] + df['ysize']
    df['Zs_end'] = df['Zs'] + df['nfrms']

    xcorr_fm = XcorrFileMatrix.from_yaml(fm.input_path)

    xcorr_fm.xcorr_options['px_size_xy'] *= args.xy_divide_by
    xcorr_fm.xcorr_options['px_size_z'] *= args.z_divide_by
    xcorr_fm.xcorr_options['max_dx'] /= args.xy_divide_by
    xcorr_fm.xcorr_options['max_dy'] /= args.xy_divide_by
    xcorr_fm.xcorr_options['max_dz'] /= args.z_divide_by
    xcorr_fm.xcorr_options['overlap_h'] /= args.xy_divide_by
    xcorr_fm.xcorr_options['overlap_v'] /= args.xy_divide_by
    xcorr_fm.xcorr_options['z_stride'] /= args.z_divide_by

    fm.save_to_yaml(args.output_file, 'w')

    with open(args.input_file, 'r') as f:
        y = yaml.load(f)

    with open(args.output_file, 'a') as f:
        yaml.dump(
            {
                'fuser-options': y['fuser-options'],
                'xcorr-options': xcorr_fm.xcorr_options,
            }, f, default_flow_style=False)

    with open(args.output_file, 'r') as f:
        file_content = f.read()

    ext = args.ext
    if not ext.startswith('.'):
        ext = '.' + ext
    old_ext = Path(df.index[0]).suffix
    if old_ext:
        file_content = file_content.replace(old_ext, ext)

    with open(args.output_file, 'w') as f:
        f.write(file_content)

    vfv = VirtualFusedVolume(args.output_file)
    logger.info(f'final shape: {vfv.shape}')


if __name__ == '__main__':
    main()



"""
This module provides `ZipWrapper`, a class for accessing a stack of images
stored in a .zip file one frame per file. Files inside the .zip archive must
will be accessed in alphabetical order.

For faster access when using compressed formats (such as JPEG2000), a cache can
optionally be anabled using `set_cache()`.
"""

import gc
import sys
import ctypes
import zipfile
import concurrent.futures
from pathlib import Path
from cachetools import LRUCache

import imageio
import numpy as np

from zetastitcher.io.inputfile_mixin import InputFileMixin

# disable cache
_cache = LRUCache(maxsize=0)
_cache.hits = 0
_cache.misses = 0


def set_cache(cache):
    """
    Set a cache for `ZipWrapper.zslice`.

    Example:

    .. code-block:: python

       import zetastitcher.io.zipwrapper as zw
       from cachetools import LRUCache
       zw.set_cache(LRUCache(maxsize=32))

    Parameters
    ----------
    cache : `cachetools:cachetools.Cache`
        The cache instance to use.
    """
    global _cache
    _cache = cache
    _cache.hits = 0
    _cache.misses = 0


def get_typecodes():
    ct = ctypes
    simple_types = [
        ct.c_byte, ct.c_short, ct.c_int, ct.c_long, ct.c_longlong,
        ct.c_ubyte, ct.c_ushort, ct.c_uint, ct.c_ulong, ct.c_ulonglong,
        ct.c_float, ct.c_double,
    ]

    return {np.dtype(ctype).str: ctype for ctype in simple_types}


def imread_wrapper(fname, internal_fname, dtype=None):
    zf = zipfile.ZipFile(str(fname), mode='r')
    a = imageio.imread(zf.read(internal_fname))
    if dtype is not None:
        a = a.astype(dtype)
    return a


class ZipWrapper(InputFileMixin):
    def __init__(self, file_path=None):
        super().__init__()
        self.file_path = file_path

        self.zf = None
        self.names = None

        if self.file_path is not None:
            self.file_path = Path(self.file_path)
            self.open()

    def open(self, file_path=None):
        if file_path is not None:
            self.file_path = Path(file_path)

        self.zf = zipfile.ZipFile(str(self.file_path), mode='r')
        setattr(self, 'close', getattr(self.zf, 'close'))
        names = self.zf.namelist()
        names.sort()

        im = imread_wrapper(self.file_path, names[0])

        self.xsize = im.shape[-1]
        self.ysize = im.shape[-2]
        self.nfrms = len(names)
        self.dtype = im.dtype

        if len(im.shape) > 2:
            self.nchannels = im.shape[0]

        self.names = names

    def frame(self, index, dtype=None, copy=None):
        a = imageio.imread(self.zf.read(self.names[index]))

        if dtype is not None:
            a = a.astype(dtype)
        return a

    def zslice(self, arg1, arg2=None, step=None, dtype=None, copy=None):
        if dtype is None:
            dtype = self.dtype

        s = list(self.shape)
        zlist = list(self._args_to_range(arg1, arg2, step))
        s[0] = len(zlist)

        out = np.zeros(s, dtype)

        # NEW ##################################

        my_futures = []

        e = concurrent.futures.ProcessPoolExecutor()

        for i, z in zip(range(s[0]), zlist):
            cache_key = f'{self.file_path}__{self.names[z]}'
            cached = _cache.get(cache_key)

            if cached is not None:
                _cache.hits += 1
                out[i] = cached
            else:
                _cache.misses += 1
                fut = e.submit(imread_wrapper, self.file_path, self.names[z], dtype)
                my_futures.append((i, cache_key, fut))

        for z, key, fut in my_futures:
            if _cache.maxsize > 0:
                _cache[key] = fut.result(None)
            out[z] = fut.result(None)

        # force release of shared memory for Python < 3.8
        if sys.version_info < (3, 8):
            import multiprocessing.heap as mph
            mph.BufferWrapper._heap = mph.Heap()
            gc.collect()

        return out



from pathlib import Path

import numpy as np

try:
    import dcimg
except ImportError:
    pass

from .ffmpeg_wrapper import FFMPEGWrapper
from .tiffwrapper import TiffWrapper
from tifffile import TiffFileError
from .zipwrapper import ZipWrapper
from .mhdwrapper import MHDWrapper

from zipfile import BadZipFile

from zetastitcher.io.inputfile_mixin import InputFileMixin


class InputFile(InputFileMixin):
    def __init__(self, file_path=None):
        super().__init__()
        self.file_path = file_path
        self.wrapper = None
        self._channel = None
        self.squeeze = True

        self.nfrms = None

        if file_path is not None:
            self.file_path = Path(self.file_path)
            self.open()

    def __repr__(self):
        return '<InputFile file_path="{}" shape={} dtype={}>'.format(
            self.file_path, self.shape, self.dtype)

    def __getitem__(self, item):
        item = np.index_exp[item]  # ensure item is a tuple
        myitem = list(item)

        # ensure all items are slice objects
        # myitem = []
        if isinstance(item[0], int):
            start = item[0]
            stop = item[0] + 1
            step = None
        elif item[0] is Ellipsis:
            start = None
            stop = self.shape[0]
            step = None
        elif isinstance(item[0], slice):
            start = item[0].start
            stop = item[0].stop
            step = item[0].step
        else:
            raise TypeError("Invalid type: {}".format(type(item[0])))

        a = self.zslice(start, stop, step)

        if not a.size:
            return a

        if item[0] is not Ellipsis:
            myitem[0] = slice(None)

        a = a[tuple(myitem)]

        if self.squeeze:
            a = np.squeeze(a)
        return a

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value):
        if not self.nchannels > 1:
            return
        self._channel = value

    @property
    def file(self):
        return self.wrapper

    @file.setter
    def file(self, value):
        self.wrapper = value
        self._setattrs()

    @property
    def shape(self):
        """Shape of the whole image stack.

        Returns
        -------
        tuple
            (:attr:`nfrms`, `channels`, :attr:`ysize`, :attr:`xsize`) where
            `channels` is the number of color channels in the image. If
            :attr:`channel` is set or if there is only one channel,
            the `channels` dimension is squeezed.
        """
        s = [self.nfrms, self.nchannels, self.ysize, self.xsize]
        if self.nchannels == 1 or self.channel is not None:
            del s[1]
        return tuple(s)

    @property
    def file_size(self):
        return self.path.stat().st_size

    def open(self, file_path=None):
        if file_path is not None:
            self.file_path = Path(file_path)

        self.path = Path(self.file_path)

        self._open()
        self._setattrs()

    def _open(self):
        if not self.path.exists():
            raise FileNotFoundError(self.path)

        try:
            self.wrapper = TiffWrapper(self.file_path)
            return
        except (ValueError, IndexError, TiffFileError):
            pass

        try:
            self.wrapper = dcimg.DCIMGFile(self.file_path)
            return
        except (NameError, ValueError, IsADirectoryError):
            pass

        try:
            self.wrapper = ZipWrapper(self.file_path)
            return
        except (AttributeError, NameError, IsADirectoryError, BadZipFile):
            pass

        try:
            self.wrapper = MHDWrapper(self.file_path)
            return
        except (ValueError, IndexError, IsADirectoryError):
            pass

        try:
            self.wrapper = FFMPEGWrapper(self.file_path)
            return
        except (ValueError, FileNotFoundError):
            pass

        from .pims_wrapper import PimsWrapper
        try:
            self.wrapper = PimsWrapper(self.file_path)
            return
        except:
            pass

        raise ValueError('Unsupported file type')

    def _setattrs(self):
        l = ['nfrms', 'xsize', 'ysize', 'nchannels', 'dtype', 'close',
             '__getitem__']

        for a in l:
            try:
                setattr(self, a, getattr(self.wrapper, a))
            except AttributeError:
                pass

    def zslice(self, arg1, arg2=None, step=None, dtype=None, copy=True):
        """Return a slice, i.e. a substack of frames.

        Parameters
        ----------
        arg1 : int
            Mandatory argument, will be passed to `python:slice`
            If arg2 and step are both None, it will be passed as `slice(arg1)`,
            i.e. it would act as the stop argument.
        arg2 : int
            If not null, will be passed as `slice(arg1, arg2, step)`
        step : int
            If not null, will be passed as `slice(arg1, arg2, step)`
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array of the original type or of `dtype`, if specified. The
            shape of the array is (`end_frame` - `start_frame`,
            :attr:`channels` :attr:`ysize`, :attr:`xsize`, :attr:`channels`)
            where `channels` is the number of color channels in the image. If
            :attr:`channel` is set or if there is only one channel, the
            `channels` dimension is squeezed.
        """
        myslice = slice(arg1, arg2, step)
        if arg2 is None and step is None:
            myslice = slice(arg1)

        if dtype is None:
            dtype = self.dtype

        ok = False
        try:
            ok = callable(getattr(self.wrapper, 'zslice'))
        except AttributeError:
            pass

        if ok:
            a = self.wrapper.zslice(myslice.start, myslice.stop, myslice.step, dtype, copy)
        else:
            myrange = list(self._args_to_range(arg1, arg2, step))
            s = list(self.shape)
            s[0] = len(myrange)
            a = np.zeros(s, dtype=dtype)
            z = 0
            for i in myrange:
                a[z] = self.wrapper.frame(i)
                z += 1

        if self.channel is not None:
            a = a[..., self.channel]
        elif self.nchannels > 1 and a.ndim >= 3:
            a = np.moveaxis(a, -1, -3)

        if len(a.shape) < len(self.shape):
            a = np.expand_dims(a, axis=0)  # add dummy Z dimension

        return a

    def zslice_idx(self, index, frames_per_slice=1, dtype=None, copy=True):
        """Return a slice, i.e. a substack of frames, by index.

        Parameters
        ----------
        index : int
            slice index
        frames_per_slice : int
            number of frames per slice
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array, see :func:`slice`.
        """
        start_frame = index * frames_per_slice
        end_frame = start_frame + frames_per_slice
        return self.zslice(start_frame, end_frame, 1, dtype, copy)

    def whole(self, dtype=None, copy=True):
        """Convenience function to retrieve the whole stack.

        Equivalent to call :func:`slice_idx` with `index` = 0 and
        `frames_per_slice` = :attr:`nfrms`

        Parameters
        ----------
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array, see :func:`slice`.
        """
        return self.zslice_idx(0, self.nfrms, dtype, copy)

    def frame(self, index, dtype=None, copy=True):
        """Convenience function to retrieve a single slice.

        Same as calling :func:`slice` and squeezing.

        Parameters
        ----------
        index : int
            slice index
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array, see :func:`slice`.
        """
        try:
            a = self.wrapper.frame(index, dtype=dtype, copy=copy)
        except AttributeError:
            a = self.zslice_idx(index, dtype=dtype, copy=copy)
        return np.squeeze(a)



import sys
import os.path
import logging
import argparse

import yaml
import humanize
import coloredlogs

from ..version import __version__

import numpy as np

from . import absolute_positions
from .fuse_runner import FuseRunner
from zetastitcher.align.filematrix import FileMatrix
from zetastitcher.align.xcorr_filematrix import XcorrFileMatrix

logger = logging.getLogger(__name__)
coloredlogs.install(level='INFO', fmt='%(levelname)s [%(name)s]: %(message)s')


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


ABS_MODE_NOMINAL_POSITIONS = 'nominal_positions'
ABS_MODE_MAXIMUM_SCORE = 'maximum_score'


def parse_args():
    parser = argparse.ArgumentParser(
        description='Fuse stitched tiles in a folder.',
        epilog='Author: Giacomo Mazzamuto <mazzamuto@lens.unifi.it>\n'
               'Version: {}'.format(__version__),
        formatter_class=CustomFormatter)

    parser.add_argument(
        'yml_file',
        help='.yml file produced by stitch align. It will also be used for '
             'saving absolute coordinates, unless option -w is specified. If a '
             'directory is specified instead of a file, uses a file named '
             '"stitch.yml"')

    group = parser.add_argument_group('output')
    group.add_argument('-o', type=str, dest='output_filename',
                       help='output file name. If not specified, no tiff '
                            'output is produced, only absolute coordinates '
                            'are computed.')

    group.add_argument('-w', type=str, dest='yml_out_file',
                       help='save data to a different .yml file')

    group.add_argument('-d', dest='debug', action='store_true',
                       help='overlay debug info')

    group.add_argument('-c', '--ch', type=int, dest='channel', help='channel')

    group.add_argument('--compression', type=str, default='zlib', help='int or string')

    group.add_argument('--downsample-xy', metavar='S', type=int, required=False,
                       help='downsample XY plane by factor S')

    group.add_argument('--zmin', type=float, default=0,
                       help='start frame (in your units)')
    me_group = group.add_mutually_exclusive_group()
    me_group.add_argument('--zmax', type=float,
                          help='end frame (noninclusive, in your units)')
    me_group.add_argument('--nz', type=int, help='number of z frames')

    group.add_argument('-s', dest='abs_mode', action='store_const',
                       const=ABS_MODE_NOMINAL_POSITIONS,
                       help='use nominal stage positions')

    group.add_argument('-f', action='store_true', default=False,
                       dest='force_recomputation',
                       help='force recomputation of absolute positions')

    group.add_argument('--no-global', action='store_true',
                       help='do not perform global optimization (where '
                            'applicable)')

    group = parser.add_argument_group('tile ordering (option -s only)')
    group.add_argument('--iX', action='store_true', dest='invert_x',
                       help='invert tile ordering along X')
    group.add_argument('--iY', action='store_true', dest='invert_y',
                       help='invert tile ordering along Y')

    group = parser.add_argument_group(
        'pixel size', 'If specified, the corresponding options can be '
                      'expressed in your custom units.')
    group.add_argument('--px-size-xy', type=float,
                       help='pixel size in the (X, Y) plane')
    group.add_argument('--px-size-z', type=float,
                       help='pixel size in the Z direction')

    args = parser.parse_args()

    return args


def preprocess_and_check_args(args):
    px_attrs = ['px_size_z', 'px_size_xy']
    if args.abs_mode == ABS_MODE_NOMINAL_POSITIONS:
        for a in px_attrs:
            if getattr(args, a, None) is None:
                logger.error(
                    "px sizes need to be specified when using option -s")
                sys.exit(1)

        for k in ['x', 'y']:
            setattr(args, 'ascending_tiles_' + k, not getattr(args,
                                                              'invert_' + k))
    else:
        if os.path.isdir(args.yml_file):
            args.yml_file = os.path.join(args.yml_file, 'stitch.yml')
        if not os.path.isfile(args.yml_file):
            logger.error(
                "No stitch file specified or found. Please specify input file "
                "or run with -s.")
            sys.exit(1)

        old_abs_mode = None
        # replace None args with values found in yml file
        if os.path.isfile(args.yml_file):
            with open(args.yml_file, 'r') as f:
                y = yaml.safe_load(f)
                try:
                    old_abs_mode = y['fuser-options']['abs_mode']
                except KeyError:
                    pass
            keys = ['px_size_z', 'px_size_xy', 'ascending_tiles_x',
                    'ascending_tiles_y']
            for k in keys:
                try:
                    attr = getattr(args, k)
                except AttributeError:
                    attr = None
                if attr is None:
                    try:
                        setattr(args, k, y['xcorr-options'][k])
                    except KeyError:
                        pass

        if args.abs_mode is None:
            args.abs_mode = old_abs_mode
        if args.abs_mode is None:
            args.abs_mode = ABS_MODE_MAXIMUM_SCORE

        if args.abs_mode != old_abs_mode:
            args.force_recomputation = True
        for a in px_attrs:
            if a not in args:
                setattr(args, a, 1)

    args.zmin = int(round(args.zmin / args.px_size_z))
    if args.zmax is not None:
        args.zmax = int(round(args.zmax / args.px_size_z))
    elif args.nz is not None:
        args.zmax = args.zmin + args.nz

    try:
        args.compression = int(args.compression)
    except ValueError:
        pass


def compute_absolute_positions(args, fm):
    xcorr_fm = XcorrFileMatrix.from_yaml(fm.input_path)
    xcorr_fm.aggregate_results()

    sdf = xcorr_fm.stitch_data_frame

    absolute_positions.compute_shift_vectors(fm.data_frame, sdf)

    if not args.no_global:
        absolute_positions.global_optimization(fm.data_frame, xcorr_fm)


def append_fuser_options_to_yaml(yml_out_file, args):
    with open(yml_out_file, 'r') as f:
        y = yaml.safe_load(f)
    fr_options = {}
    keys = ['px_size_xy', 'px_size_z']
    if args.abs_mode == ABS_MODE_NOMINAL_POSITIONS:
        keys += ['ascending_tiles_x', 'ascending_tiles_y']
    else:
        keys += ['abs_mode']
    for k in keys:
        fr_options[k] = getattr(args, k)
    y['fuser-options'] = fr_options

    with open(yml_out_file, 'w') as f:
        yaml.dump(y, f, default_flow_style=False)


def main():
    args = parse_args()
    preprocess_and_check_args(args)

    logger.info("invert X: {}, invert Y: {}".format(
        not args.ascending_tiles_x, not args.ascending_tiles_y))
    logger.info("voxel size (ZYX): {} * {} * {}".format(
        args.px_size_z, args.px_size_xy, args.px_size_xy))

    # =========================================================================
    # init FileMatrix
    # =========================================================================
    fm = FileMatrix(args.yml_file,
                    ascending_tiles_x=args.ascending_tiles_x,
                    ascending_tiles_y=args.ascending_tiles_y)

    if args.force_recomputation:
        fm.clear_absolute_positions()

    cols = fm.data_frame.columns

    if 'Xs' in cols and 'Ys' in cols and 'Zs' in cols:
        logger.info('using absolute positions from {}'.format(args.yml_file))
    else:
        if args.abs_mode == ABS_MODE_NOMINAL_POSITIONS:
            logger.info('absolute positions mode: using nominal stage coordinates')
            fm.compute_nominal_positions(args.px_size_z, args.px_size_xy)
            yml_out_file = args.yml_out_file
        else:
            if not os.access(args.yml_file, os.W_OK):
                raise ValueError('cannot write to {}'.format(args.yml_file))
            compute_absolute_positions(args, fm)
            yml_out_file = args.yml_file

        if yml_out_file is None:
            logger.warning('No .yml output file specified (-w option)')
        else:
            if os.path.isfile(yml_out_file):
                fm.save_to_yaml(yml_out_file, 'update')
            else:
                fm.save_to_yaml(yml_out_file, 'w')
            append_fuser_options_to_yaml(yml_out_file, args)

    # =========================================================================
    # init FuseRunner
    # =========================================================================
    fr = FuseRunner(fm)
    bytes_human = humanize.naturalsize(np.prod(fr.output_shape) * fr.dtype.itemsize, binary=True)
    logger.info(f'fused shape, whole volume: {fr.output_shape}, {bytes_human}')
    if args.output_filename is not None:

        keys = ['zmin', 'zmax', 'output_filename', 'debug', 'channel',
                'compression', 'downsample_xy']

        for k in keys:
            setattr(fr, k, getattr(args, k))

        bytes_human = humanize.naturalsize(np.prod(fr.output_shape) * fr.dtype.itemsize, binary=True)
        logger.info(f'output shape: {fr.output_shape}, {bytes_human}')

        fr.run()
    else:
        logger.warning("No output file specified.")


if __name__ == '__main__':
    main()



import os
import sys
import time
import queue
import logging
import argparse
import threading
import concurrent.futures
from datetime import timedelta

import json
import yaml

import coloredlogs

import numpy as np
import pandas as pd

from zetastitcher.io.inputfile import InputFile
from zetastitcher.align.filematrix import FileMatrix
from zetastitcher.align.xcorr_filematrix import XcorrFileMatrix
from zetastitcher.align.dog import align_dog
from zetastitcher.fuse import absolute_positions
from zetastitcher.fuse.__main__ import ABS_MODE_MAXIMUM_SCORE

from zetastitcher.version import __version__


logger = logging.getLogger(__name__)
coloredlogs.install(level='INFO', fmt='%(levelname)s [%(name)s]: %(message)s')


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def parse_args():
    parser = argparse.ArgumentParser(
        description='''
Stitch tiles in a folder.

The following naming conventions are used:
* Z is the direction along the stack height,
* (X, Y) is the frame plane,
* y is the direction along which frames are supposed to overlap,
* x is the direction orthogonal to y in the frame plane (X, Y).

Unless otherwise stated, all values are expected in px.
    ''',
        epilog='Author: Giacomo Mazzamuto <mazzamuto@lens.unifi.it>\n'
               'Version: {}'.format(__version__),
        formatter_class=CustomFormatter)

    parser.add_argument('input_folder', help='input folder')
    parser.add_argument('-o', type=str, default='stitch.yml', dest='output_file', help='output file')
    parser.add_argument('-c', '--ch', type=int, dest='channel', help='color channel')
    parser.add_argument('-j', type=int, dest='n_of_workers',
                        help='number of parallel jobs (defaults to number of system cores)')
    parser.add_argument('-r', action='store_true', dest='recursive', help='recursively look for files')
    parser.add_argument('-e', action='store_true', dest='equal_shape',
                        help='consider tiles of identical shape (results in slightly faster loading)')

    group = parser.add_argument_group(
        'pixel size', 'If specified, the corresponding options can be '
                      'expressed in your custom units.')
    group.add_argument('--px-size-xy', type=float, default=1,
                       help='pixel size in the (X, Y) plane')
    group.add_argument('--px-size-z', type=float, default=1,
                       help='pixel size in the Z direction')

    group = parser.add_argument_group('maximum shifts')
    group.add_argument('--dz', type=float, required=True, dest='max_dz',
                       help='maximum allowed shift along Z')

    group.add_argument('--dy', type=float, required=True, dest='max_dy',
                       help='maximum allowed shift along y (the stitching '
                            'axis) relative to the nominal overlap')

    group.add_argument('--dx', type=float, required=True, dest='max_dx',
                       help='maximum allowed shift along x (lateral shift)')

    group = parser.add_argument_group('overlaps')
    group.add_argument('--overlap', type=float, help='nominal overlap, H & V')
    group.add_argument('--overlap-h', type=float, metavar='OH',
                       help='nominal overlap along the horizontal axis')

    group.add_argument('--overlap-v', type=float, metavar='OV',
                       help='nominal overlap along the vertical axis')

    group = parser.add_argument_group(
        'multiple sampling along Z',
        description='Measure the optimal shift at different heights around '
                    'the center of the stack')

    group.add_argument('--z-samples', type=int, default=1, metavar='ZSAMP',
                       help='number of samples to take along Z')

    group.add_argument('--z-stride', type=float, default=None,
                       help='stride used for multiple Z sampling')

    group = parser.add_argument_group('tile ordering')
    group.add_argument('--iX', action='store_true', dest='invert_x',
                       help='invert tile ordering along X')
    group.add_argument('--iY', action='store_true', dest='invert_y',
                       help='invert tile ordering along Y')

    args = parser.parse_args()

    if args.overlap is None:
        if args.overlap_h is None or args.overlap_v is None:
            logger.error('Missing overlap options. '
                         'Use --overlap, or both --overlap-h and --overlap-v')
            sys.exit(1)
    elif args.overlap_h is not None or args.overlap_v is not None:
        logger.error('Incompatible options: --overlap and --overlap-h, '
                     '--overlap-v')
        sys.exit(1)
    else:
        setattr(args, 'overlap_h', args.overlap)
        setattr(args, 'overlap_v', args.overlap)

    args.max_dx = int(round(args.max_dx / args.px_size_xy))
    args.max_dy = int(round(args.max_dy / args.px_size_xy))
    args.max_dz = int(round(args.max_dz / args.px_size_z))
    if args.z_stride is not None:
        args.z_stride = int(round(args.z_stride / args.px_size_z))

    args.overlap_v = int(round(args.overlap_v / args.px_size_xy))
    args.overlap_h = int(round(args.overlap_h / args.px_size_xy))

    if args.max_dy > args.overlap_h or args.max_dy > args.overlap_v:
        raise ValueError("dy can't be larger than overlap")

    setattr(args, 'ascending_tiles_x', not args.invert_x)
    setattr(args, 'ascending_tiles_y', not args.invert_y)

    return args


def worker(item, overlap_dict, channel, max_dz, max_dy, max_dx):
    aname = item['aname']
    bname = item['bname']
    z_frame = item['z_frame']
    axis = item['axis']
    overlap = overlap_dict[axis]

    a = InputFile(aname)
    b = InputFile(bname)

    z_min = z_frame - max_dz
    z_max = z_frame + max_dz + 1

    aslice = a.zslice(z_min, z_max, copy=True)
    if a.nchannels > 1:
        if channel is not None:
            aslice = aslice[:, channel]
        else:
            aslice = np.sum(aslice.astype(np.float32), axis=1)
    if axis == 2:
        aslice = np.rot90(aslice, axes=(-1, -2))
    aslice = aslice[..., -(overlap + max_dy):, :]

    bframe = b.zslice_idx(z_frame, copy=True)
    if b.nchannels > 1:
        if channel is not None:
            bframe = bframe[:, channel]
        else:
            bframe = np.sum(bframe.astype(np.float32), axis=1)
    if axis == 2:
        bframe = np.rot90(bframe, axes=(-1, -2))
    bframe = bframe[..., :overlap - max_dy, :]

    aslice = aslice.astype(np.float32)
    bframe = bframe.astype(np.float32)

    output_shape = np.array(aslice.shape) + np.array((0, 0, 2 * max_dx)) - np.array(bframe.shape) + 1
    output_shape[0] = aslice.shape[0]
    xcorr = np.zeros(output_shape)

    for i in range(xcorr.shape[0]):
        cc, max_loc = align_dog(aslice[i], bframe[0], 0, max_dx)
        xcorr[i] = cc

    shift = list(np.unravel_index(np.argmax(xcorr), xcorr.shape))
    score = xcorr[tuple(shift)]
    if score < 0 or score > 1:
        score = 0

    item['score'] = score
    item['dz'] = shift[0]
    item['dy'] = shift[1]
    item['dx'] = shift[2]

    return item


class Runner(object):
    def __init__(self):
        self.channel = None
        self.processing_list = None
        self.fut_q = None
        self.output_q = None
        self.input_folder = None
        self.output_file = None
        self.z_samples = None
        self.z_stride = None
        self.overlap_v = None
        self.overlap_h = None
        self.max_dx = None
        self.max_dy = None
        self.max_dz = None
        self.ascending_tiles_x = True
        self.ascending_tiles_y = True
        self.recursive = False
        self.equal_shape = False
        self.df = None
        self.fm = None
        self.px_size_xy = 1
        self.px_size_z = 1
        self.n_of_workers = None

    @property
    def overlap_dict(self):
        return {1: self.overlap_v, 2: self.overlap_h}

    def initialize_list(self):
        fm = FileMatrix(self.input_folder, self.ascending_tiles_x, self.ascending_tiles_y,
                        recursive=self.recursive, equal_shape=self.equal_shape)
        self.fm = fm

        if self.z_samples > 1 and self.z_stride is None:
            self.z_stride = ((fm.data_frame.iloc[0].nfrms - self.max_dz * self.z_samples) // self.z_samples).item()

        if self.z_stride is None:
            self.z_stride = 0

        stitch_X = {
            'axis': 2,
            'sort': ['Z', 'X', 'Y'],
            'groupby': 'Y'
        }

        stitch_Y = {
            'axis': 1,
            'sort': ['Z', 'Y', 'X'],
            'groupby': 'X'
        }

        mylist = []

        for s in fm.slices():
            df = fm.data_frame.loc[list(s.nodes())]
            for stitch_config in [stitch_X, stitch_Y]:
                view = df.sort_values(stitch_config['sort'],
                                      ascending=True)
                view = view.groupby(stitch_config['groupby'])

                for name, group in view:
                    Z_from = int(max(group['Z']))
                    Z_to = int(min(group['Z_end']))
                    tiles = list(group.itertuples())
                    atile = tiles[0]
                    for btile in tiles[1::]:
                        central_frame = (Z_to - Z_from) // 2
                        start_frame = (
                            central_frame
                            - (self.z_samples // 2 * self.z_stride)
                            + (0 if self.z_samples % 2 else self.z_stride // 2)
                        )
                        for i in range(0, self.z_samples):
                            mylist.append({
                                'aname': atile.Index,
                                'bname': btile.Index,
                                'z_frame': start_frame + i * self.z_stride,
                                'axis': stitch_config['axis'],
                            })
                        atile = btile
        self.processing_list = mylist

    def keep_filling_fut_queue(self):
        e = concurrent.futures.ProcessPoolExecutor(max_workers=self.n_of_workers)

        for item in self.processing_list:
            self.fut_q.put(e.submit(worker, item, self.overlap_dict, self.channel, self.max_dz, self.max_dy, self.max_dx))

        self.fut_q.put(None)

    def output_worker(self):
        i = 1
        while True:
            fut = self.fut_q.get()
            if fut is None:
                self.fut_q.task_done()
                break

            item = fut.result()

            progress = 100 * i / len(self.processing_list)
            aname = item['aname']
            bname = item['bname']
            z_frame = item['z_frame']
            shift = [item['dz'], item['dy'], item['dx']]
            score = item['score']
            logger.info(f'{progress:.0f}%\t{aname}\t{bname}\t{z_frame}\t{shift}\t{score:.3f}')

            self.output_q.put(item)
            self.fut_q.task_done()

            i += 1

    def run(self):
        out_dir = os.path.dirname(os.path.abspath(self.output_file))
        if not os.access(out_dir, os.W_OK):
            raise ValueError('cannot write to {}'.format(self.output_file))

        self.initialize_list()
        self.fut_q = queue.Queue()
        self.output_q = queue.Queue()

        t = threading.Thread(target=self.output_worker)
        t.start()

        self.keep_filling_fut_queue()

        # block until all tasks are done
        self.fut_q.join()

        df = pd.DataFrame(self.output_q.queue)
        self.df = df

        xcorr_fm = XcorrFileMatrix.from_data(self.xcorr_options, self.df)
        xcorr_fm.aggregate_results()

        sdf = xcorr_fm.stitch_data_frame

        absolute_positions.compute_shift_vectors(self.fm.data_frame, sdf)
        absolute_positions.global_optimization(self.fm.data_frame, xcorr_fm)

        self.save_results_to_file()

        cols = ['score', 'dz', 'dy', 'dx']
        print(df[cols].describe())

    @property
    def xcorr_options(self):
        attrs = ['max_dx', 'max_dy', 'max_dz', 'overlap_v', 'overlap_h',
                 'ascending_tiles_x', 'ascending_tiles_y', 'px_size_xy',
                 'px_size_z', 'z_samples', 'z_stride']

        options = {}
        for attr in attrs:
            options[attr] = getattr(self, attr)

        return options

    def save_results_to_file(self):
        self.fm.save_to_yaml(self.output_file, 'w')

        with open(self.output_file, 'a') as f:
            yaml.dump(
                {
                    'xcorr-options': self.xcorr_options,
                    'xcorr': json.loads(self.df.to_json(orient='records')),
                    'fuser-options': {'abs_mode': ABS_MODE_MAXIMUM_SCORE},
                }, f, default_flow_style=False)


def main():
    arg = parse_args()

    r = Runner()

    keys = ['input_folder', 'output_file', 'channel', 'max_dx', 'max_dy',
            'max_dz', 'z_samples', 'z_stride', 'overlap_v', 'overlap_h',
            'ascending_tiles_x', 'ascending_tiles_y', 'px_size_xy',
            'px_size_z', 'n_of_workers', 'recursive', 'equal_shape']

    for key in keys:
        setattr(r, key, getattr(arg, key))

    t = time.time()
    r.run()
    elapsed = timedelta(seconds=time.time() - t)
    logger.info(f'elapsed  time: {elapsed}')


if __name__ == '__main__':
    main()



from .mapper import Term, QPBuilder

def pretty_solver_debug(solver_matrices, variables):
    builder = QPBuilder()
    objective = builder.create_expression()
    for i, vi in enumerate(variables):
        for j, vj in enumerate(variables):
            objective.add_term(
                term=Term([vi, vj]),
                coeff=solver_matrices.P[i,j]
            )

    for i, vi in enumerate(variables):
        objective.add_term(
            term=Term([vi]),
            coeff=solver_matrices.q[i]
        )
    print('DEBUG MAT OBJ', str(objective))


import cv2 as cv

import numpy as np


def twoD_gaussian_kernel(ksize, sigma):
    filter = cv.getGaussianKernel(ksize, sigma)
    return cv.mulTransposed(filter, False)


def dog(image, ksize=100, sigma1=5, sigma2=10):
    filter1 = twoD_gaussian_kernel(ksize, sigma1)
    filter2 = twoD_gaussian_kernel(ksize, sigma2)
    cv.normalize(filter1, filter1, 1, 0, cv.NORM_L1)
    cv.normalize(filter2, filter2, 1, 0, cv.NORM_L1)
    filter = filter1 - filter2

    return cv.filter2D(image, -1, filter)


def crossCorr(image1, image2, padding_y, padding_x):
    temp1 = image1.astype(np.float32)
    temp2 = image2.astype(np.float32)

    padding = (padding_y, padding_y, padding_x, padding_x)

    padded = cv.copyMakeBorder(temp1, *padding, cv.BORDER_CONSTANT)

    return cv.matchTemplate(padded, temp2, cv.TM_CCORR_NORMED)


def align_dog(i1, i2, padding_y, padding_x):
    dog1 = dog(i1)
    dog2 = dog(i2)

    cc = crossCorr(dog1, dog2, padding_y, padding_x)

    min_val, max_val, min_loc, max_loc = cv.minMaxLoc(cc)
    return cc, max_loc



from collections import defaultdict
from scipy.sparse import coo_matrix
import numpy as np

def linear_expr(variables, coefficients, const_coeff=0.):
    var2coeff = {v:c for v, c in zip(variables, coefficients)}
    const_coeff = const_coeff
    return LinearExpr(var2coeff, const_coeff)


class LinearExpr(object):
    def __init__(self, var2coeff, const_coeff):
        self.var2coeff = var2coeff
        self.const_coeff = const_coeff
    
    def variables_set(self):
        return set(self.var2coeff)
    
    def __add__(self, other):
        res = defaultdict(float)
        res.update(self.var2coeff)
        for v, c in other.var2coeff.items():
            res[v] += c
        return LinearExpr(var2coeff=dict(res), const_coeff=self.const_coeff + other.const_coeff)
        
    def __sub__(self, other):
        return self + (-1 * other)
    
    def __mul__(self, scalar):
        return LinearExpr(var2coeff={v:c*scalar for v, c in self.var2coeff.items()} ,const_coeff=self.const_coeff*scalar)
    
    def __rmul__(self, scalar):
        return self.__mul__(scalar)
    
    def __repr__(self):
        if self.const_coeff:
            l = ["{}".format(self.const_coeff)]
        else:
            l = []
        l += ["{:+} * {}".format(c, v) for v, c in self.var2coeff.items()]
        return " ".join(l)

def multidim_lin_expr(variables, A, b):
    n, m = A.shape
    assert b.shape == (n,)
    assert len(variables) == n
    lin_expr_list = []
    for i in range(n):
        lin_expr = linear_expr(variables, coefficients=A[i,:], const_coeff=b[i])
        lin_expr_list.append(lin_expr)
    return MultiDimLinearExpr(lin_expr_list)


class MultiDimLinearExpr(object):
    def __init__(self, lin_expr_list):
        self.lin_expr_list = lin_expr_list
        
    def variables_set(self):
        res = set()
        for lin_expr in self.lin_expr_list:
            res = res.union(lin_expr.variables_set())
        return res
        
    def __iter__(self):
        yield from iter(self.lin_expr_list)
    
    def __len__(self):
        return len(self.lin_expr_list)
    
    def extend(self, other):
        for lin_expr in other:
            self.lin_expr_list.append(lin_expr)
    
    def add(self, other):
        if self.lin_expr_list:
            self_len = len(self.lin_expr_list)
            other_len = len(other.lin_expr_list)

            if self_len != other_len:
                raise ValueError('programming error dimensions do not match. {} != {}'.format(self_len, other_len))
            
            lin_expr_list = [expr1 + expr2
                for expr1, expr2 in zip(self.lin_expr_list, other.lin_expr_list)]
            return MultiDimLinearExpr(lin_expr_list)
        else:
            return other # XXX potentially unsafe
        
    def append(self, lin_expr):
        self.lin_expr_list.append(lin_expr)
    
    def get_vars_and_matrices(self):
        n_eq = len(self.lin_expr_list)
        variables = sorted(self.variables_set())
        n_vars = len(variables)
        variable2index = {v:idx for idx, v in enumerate(variables)}
        A_i = []
        A_j = []
        A_data = []
        b = np.zeros((n_eq,))
        for i, lin_expr in enumerate(self.lin_expr_list):
            b[i] += lin_expr.const_coeff
            for v, coeff in lin_expr.var2coeff.items():
                j = variable2index[v]
                A_i.append(i)
                A_j.append(j)
                A_data.append(coeff)
        A = coo_matrix((A_data, (A_i, A_j)), shape=(n_eq, n_vars))
        A.eliminate_zeros()
        return variables, A, b

def main():
    expr1 = linear_expr(variables=['a', 'b'], coefficients=[0.1, -0.4])
    expr2 = linear_expr(variables=['b', 'c'], coefficients=[0.1, -4])
    print(expr1)
    print(expr2)
    print(expr1 + expr2)
    print(expr1 - expr2)

if __name__ == '__main__':
    main()




import numpy as np


def ConstraintTuple(ndims, eps_reg=None):
    class ConstraintTupleND(object):
        __slots__ = ['v', 'w', 'Lambda', 'p', 'ub', 'lb']
        def __init__(self, v, w, Lambda, p, ub=None, lb=None):
            self.v = v
            self.w = w
            self.Lambda = Lambda
            self.p = p
            self.ub = ub
            self.lb = lb
            self._check_dims(Lambda, p, ub, lb)
            if eps_reg is not None:
                if eps_reg < 0:
                    raise ValueError('eps_reg < 0')
                self.Lambda += eps_reg * np.eye(Lambda.shape[0])

        def _check_dims(self, Lambda, p, ub, lb):
            if Lambda.shape != (ndims, ndims):
                raise ValueError('Lambda.shape should be {} rather than {}.'.format((ndims, ndims), Lambda.shape))
            if p.shape != (ndims,):
                raise ValueError('p.shape should be {} rather than {}.'.format((ndims,), p.shape))

            if ub is not None and ub.shape != (ndims,):
                raise ValueError('ub.shape should be {} rather than {}.'.format((ndims,), ub.shape))

            if lb is not None and lb.shape != (ndims,):
                raise ValueError('lb.shape should be {} rather than {}.'.format((ndims,), lb.shape))
    return ConstraintTupleND



import numpy as np


class InputFileMixin:
    def __init__(self):
        self.path = None
        self.xsize = None
        self.ysize = None
        self.nfrms = None
        self.dtype = None
        self.nchannels = 1

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        pass

    @property
    def shape(self):
        """Shape of the whole image stack.

        Returns
        -------
        tuple
            (:attr:`nfrms`, :attr:`ysize`, :attr:`xsize`, `channels`) where
            `channels` is the number of color channels in the image. If
            :attr:`channel` is set or if there is only one channel,
            the `channels` dimension is squeezed.
        """
        s = [self.nfrms, self.ysize, self.xsize, self.nchannels]
        if self.nchannels == 1:
            s = s[:-1]
        return tuple(s)

    @property
    def file_size(self):
        return self.path.stat().st_size

    @property
    def array_size(self):
        """Size in bytes of the whole virtual array"""
        return np.prod(self.shape).item() * self.dtype.itemsize

    @property
    def frame_size(self):
        """Size in bytes of a single frame"""
        return self.xsize * self.ysize * self.nchannels * self.dtype.itemsize

    def _normalize_slice(self, myslice):
        """
        Return the `python:slice` that would return the same elements but using
        only positive numbers in `start`, `stop` and `step`.
        """
        if myslice is Ellipsis:
            myslice = slice(None)
        start = myslice.start
        stop = myslice.stop
        step = 1 if myslice.step is None else myslice.step

        curr_max = self.nfrms

        if start and start < 0:
            start += curr_max

        if stop and stop < 0:
            stop += curr_max

        if step > 0:
            if start is None:
                start = 0
            if stop is None:
                stop = curr_max
        else:
            step *= -1

            if start is None:
                start = curr_max

            if stop is None:
                stop = -1

            mod = (start - stop) % step

            temp = start
            start = stop
            stop = temp

            start += mod if mod else 1
            stop += 1

        if stop > curr_max:
            stop = curr_max

        return slice(start, stop, step)

    @staticmethod
    def _args_to_slice(arg1, arg2=None, step=None):
        myslice = slice(arg1, arg2, step)
        if arg2 is None and step is None:
            myslice = slice(arg1)
        return myslice

    @staticmethod
    def _args_to_range(arg1, arg2=None, step=None):
        myrange = range(0)
        myslice = InputFileMixin._args_to_slice(arg1, arg2, step)
        if myslice.stop:
            myrange = range(myslice.stop)
        if myslice.start and myslice.stop:
            myrange = range(myslice.start, myslice.stop)
        if myslice.start and myslice.stop and myslice.step:
            myrange = range(myslice.start, myslice.stop, myslice.step)
        return myrange

    @staticmethod
    def _slice_to_range(myslice):
        return InputFileMixin._args_to_range(myslice.start, myslice.stop, myslice.step)



import json
import mimetypes
import subprocess as sp
from pathlib import Path

import numpy as np

from zetastitcher.io.inputfile_mixin import InputFileMixin


class FFMPEGWrapper(InputFileMixin):
    def __init__(self, file_path=None):
        super().__init__()
        self.file_path = file_path

        self.proc = None
        self.pix_fmt = None

        self._probed_dict = None

        if file_path is not None:
            self.file_path = Path(self.file_path)
            self.open()

    @property
    def shape(self):
        s = (self.nfrms, self.ysize, self.xsize)
        if self.nchannels > 1:
            s = s + (self.nchannels,)
        return s

    def open(self, file_path=None):
        if file_path is not None:
            self.file_path = Path(file_path)

        mtype = mimetypes.guess_type(self.file_path)[0]
        if mtype is None or not mtype.startswith('video'):
            raise ValueError(f'Invalid MIME type:{mtype}')

        cmd = [
            'ffprobe',
            '-v', 'quiet',
            '-i', str(self.file_path),
            '-print_format', 'json',
            '-show_format',
            '-show_streams',
            '-select_streams', 'v:0'
        ]

        pipe = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
        self._probed_dict = json.loads(pipe.stdout.read().decode('utf-8'))
        pipe.communicate()

        if len(self._probed_dict) == 0 \
            or 'mp4' not in self._probed_dict['format']['format_name']:
                raise ValueError('Unrecognized format for FFMPEGWrapper')

        self.nfrms = int(self._probed_dict['streams'][0]['nb_frames'])
        self.xsize = self._probed_dict['streams'][0]['width']
        self.ysize = self._probed_dict['streams'][0]['height']
        pix_fmt = self._probed_dict['streams'][0]['pix_fmt']

        if 'gray' in pix_fmt or 'mono' in pix_fmt or 'pal8' in pix_fmt:
            self.nchannels = 1
        elif 'yuva' in pix_fmt:
            self.nchannels = 4
        elif 'yuv' in pix_fmt or 'rgb' in pix_fmt or 'gbr' in pix_fmt or \
                'bgr' in pix_fmt:
            self.nchannels = 3
        self.pix_fmt = pix_fmt
        self.dtype = np.dtype(np.uint8)

    def zslice(self, arg1, arg2=None, step=None, dtype=None, copy=True):
        myslice = self._args_to_slice(arg1, arg2, step)
        norm_slice = self._normalize_slice(myslice)

        if norm_slice.start >= norm_slice.stop:
            return np.array([])

        command = [
            'ffmpeg',
            '-i', str(self.file_path),
            '-f', 'image2pipe',
            '-vcodec', 'rawvideo',
            '-pix_fmt', 'gray' if 'gray' in self.pix_fmt else 'rgb24',
            '-'
        ]
        pipe = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10**8)

        i = 0
        frame_shape = (1, self.ysize, self.xsize, self.nchannels)
        frame_bytes = np.prod(frame_shape) * self.dtype.itemsize

        while i < norm_slice.start:  # skip initial frames
            pipe.stdout.read(frame_bytes)
            i += 1

        shape = list(frame_shape)
        shape[0] = len(self._slice_to_range(norm_slice))

        a = np.zeros(shape, dtype=self.dtype)

        z = 0
        while i < norm_slice.stop:
            raw = pipe.stdout.read(frame_bytes)
            a[z] = np.frombuffer(raw, dtype=self.dtype).reshape(frame_shape)[0]
            for _ in range(norm_slice.step - 1):  # skip frames
                pipe.stdout.read(frame_bytes)
            i += norm_slice.step
            z += 1

        pipe.communicate()

        if myslice.step and myslice.step < 0:
            a = a[::-1]

        a = np.squeeze(a)

        if dtype is None:
            return a
        return a.astype(dtype)
